from __future__ import absolute_import

from datetime import datetime, timedelta

import requests
from six import string_types

from ...utilities import timezone
from ..exceptions import InvalidArgument

try:
    # noinspection PyUnresolvedReferences
    from typing import Union, Any
except ImportError:
    pass


_valid_identifier_characters = set(
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-._"
)
_MIN_ID_LENGTH = 1
_MAX_ID_LENGTH = 1024

_allowed_file_types = {
    # PDFs
    "pdf",
    # Word documents
    "docx",
    "docm",
    "doc",
    "rtf",
    # PowerPoint presentations
    "pptx",
    "pptm",
    "ppt",
}


def validate_identifier(identifier):
    # type: (str) -> str
    if not identifier:
        raise InvalidArgument("identifier", "`identifier` cannot be empty.")
    if not isinstance(identifier, string_types):
        raise InvalidArgument(
            "identifier", "`identifier` must be a string (of type `str`)."
        )
    if not _MIN_ID_LENGTH <= len(identifier) <= _MAX_ID_LENGTH:
        raise InvalidArgument(
            "identifier",
            "`identifier` must be between {min} and {max} characters long.".format(
                min=_MIN_ID_LENGTH, max=_MAX_ID_LENGTH
            ),
        )
    if not all(c in _valid_identifier_characters for c in identifier):
        raise InvalidArgument(
            "identifier",
            '`identifier` must only contain ASCII letters, numbers, and the characters "-._".',
        )
    return identifier


def validate_file_type(file_type):
    # type: (str) -> str
    if not file_type:
        raise InvalidArgument("file_type", "`file_type` cannot be empty.")
    munged_file_type = str(file_type).lower()
    if munged_file_type not in _allowed_file_types:
        raise InvalidArgument(
            "file_type", '"{}" is not a valid file type'.format(file_type)
        )
    return munged_file_type


def validate_file(file):
    # type: (Any) -> Any
    if not file:
        raise InvalidArgument("file", "`file` must be specified.")
    if any(not hasattr(file, attr) for attr in ("mode", "read")):
        raise InvalidArgument(
            "file",
            "the given object doesn't appear to be a file - it's missing properties a file object should have.",
        )
    if hasattr(file, "readable") and not file.readable():
        raise InvalidArgument("file", "`file.readable()` is False.")
    if "b" not in file.mode:
        raise InvalidArgument(
            "file",
            "the given file doesn't appear to be open in binary mode. It's mode is '{}'.".format(
                file.mode
            ),
        )
    return file


def validate_url(url):
    # type: (str) -> str
    if not url:
        raise InvalidArgument("url", "`url` cannot be empty.")
    if not isinstance(url, string_types):
        raise InvalidArgument("url", "`url` must be a string (of type `str`).")
    if len(url) > 2048:
        raise InvalidArgument("url", "`url` must be no longer than 2048 characters.")
    try:
        scheme, netloc, _, _, _, _ = requests.utils.urlparse(url)
    except Exception as ex:
        wrapper = InvalidArgument("url", "the value could not be parsed as a URL.")
        wrapper.__cause__ = ex
        raise wrapper
    if scheme not in ("http", "https"):
        raise InvalidArgument("url", '`url` must have "http" or "https" as its scheme.')
    if not netloc:
        raise InvalidArgument("url", "the given URL is malformed.")
    return url


def _validate_datetime_or_timedelta(parameter_name, value):
    # type: (Union[datetime, timedelta]) -> datetime
    if isinstance(value, timedelta):
        if value.total_seconds() <= 0:
            raise InvalidArgument(
                parameter_name,
                "if a timedelta, `{parameter}` must be positive, but it was zero or negative.".format(
                    parameter=parameter_name
                ),
            )
        return datetime.now(tz=timezone.utc) + value

    if not isinstance(value, datetime):
        raise InvalidArgument(
            parameter_name,
            "`{parameter}` must be a datetime or a timedelta.".format(
                parameter=parameter_name
            ),
        )
    # Make the datetime aware, with UTC as its timezone. If it was naive, we assume it was in UTC time.
    if value.utcoffset() is None:
        value = value.replace(tzinfo=timezone.utc)
    else:
        value = value.astimezone(timezone.utc)
    # Now we can safely check if it's in the future.
    if value < datetime.now(tz=timezone.utc):
        raise InvalidArgument(
            parameter_name,
            "`{parameter}` must be in the future.".format(parameter=parameter_name),
        )

    return value


def validate_expires(expires):
    # type: (Union[datetime, timedelta]) -> datetime
    return _validate_datetime_or_timedelta("expires", expires)


def validate_valid_until(valid_until):
    # type: (Union[datetime, timedelta]) -> datetime
    return _validate_datetime_or_timedelta("valid_until", valid_until)



import random
import string

# Constants for generating random unique (with high probability) identifiers:
_RANDOM_ID_LENGTH = 12
_RANDOM_ID_CHARSET = string.ascii_letters


def generate_identifier():
    # type: () -> str
    return "".join(
        random.choice(_RANDOM_ID_CHARSET) for _ in range(_RANDOM_ID_LENGTH)  # nosec
    )



from __future__ import absolute_import

from datetime import datetime

from ...utilities.timestamp import parse_datetime

try:
    # noinspection PyUnresolvedReferences
    from typing import Optional
except ImportError:
    pass


class ComparisonSide(object):
    def __init__(self, file_type, source_url, display_name):
        # type: (str, Optional[str], Optional[str]) -> None
        self.__file_type = file_type
        self.__source_url = source_url
        self.__display_name = display_name

    @property
    def file_type(self):
        # type: () -> str
        return self.__file_type

    @property
    def source_url(self):
        # type: () -> Optional[str]
        return self.__source_url

    @property
    def display_name(self):
        # type: () -> Optional[str]
        return self.__display_name

    def __str__(self):
        # type: () -> str
        return "{{file_type: {}{}{}}}".format(
            repr(self.file_type),
            ", source_url: {}".format(repr(self.source_url)) if self.source_url else "",
            ", display_name: {}".format(repr(self.display_name))
            if self.display_name
            else "",
        )

    def __repr__(self):
        # type: () -> str
        return "ComparisonSide(file_type={}, source_url={}, display_name={})".format(
            repr(self.file_type), repr(self.source_url), repr(self.display_name),
        )


class Comparison(object):
    def __init__(
        self,
        identifier,  # type: str
        left,  # type: ComparisonSide
        right,  # type: ComparisonSide
        public,  # type: bool
        creation_time,  # type: datetime
        expiry_time,  # type: Optional[datetime]
        ready,  # type: bool
        ready_time,  # type: Optional[datetime]
        failed,  # type: Optional[bool]
        error_message,  # type: Optional[str]
    ):
        self.__identifier = identifier
        self.__left = left
        self.__right = right
        self.__creation_time = creation_time
        self.__public = public
        self.__expiry_time = expiry_time
        self.__ready = ready
        self.__ready_time = ready_time
        self.__failed = failed
        self.__error_message = error_message

    @property
    def identifier(self):
        # type: () -> str
        return self.__identifier

    @property
    def left(self):
        # type: () -> ComparisonSide
        return self.__left

    @property
    def right(self):
        # type: () -> ComparisonSide
        return self.__right

    @property
    def public(self):
        # type: () -> bool
        return self.__public

    @property
    def creation_time(self):
        # type: () -> datetime
        return self.__creation_time

    @property
    def expiry_time(self):
        # type: () -> Optional[datetime]
        return self.__expiry_time

    @property
    def ready(self):
        # type: () -> bool
        return self.__ready

    @property
    def ready_time(self):
        # type: () -> Optional[datetime]
        return self.__ready_time

    @property
    def failed(self):
        # type: () -> Optional[bool]
        return self.__failed

    @property
    def error_message(self):
        # type: () -> Optional[str]
        return self.__error_message

    def __str__(self):
        # type: () -> str
        return "Comparison(identifier={}, left={}, right={}, public={}, creation_time={}, expiry_time={}, ready={}, ready_time={}, failed={}, error_message={})".format(
            repr(self.identifier),
            str(self.left),
            str(self.right),
            str(self.public),
            self.creation_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
            self.expiry_time.strftime("%Y-%m-%dT%H:%M:%SZ")
            if self.expiry_time
            else None,
            str(self.ready),
            self.ready_time.strftime("%Y-%m-%dT%H:%M:%SZ") if self.ready_time else None,
            str(self.failed),
            "<{} chars>".format(len(self.error_message))
            if self.error_message is not None
            else None,
        )

    def __repr__(self):
        # type: () -> str
        return "Comparison(identifier={}, left={}, right={}, public={}, creation_time={}, expiry_time={}, ready={}, ready_time={}, failed={}, error_message={})".format(
            repr(self.identifier),
            repr(self.left),
            repr(self.right),
            repr(self.public),
            repr(self.creation_time),
            repr(self.expiry_time),
            repr(self.ready),
            repr(self.ready_time),
            repr(self.failed),
            repr(self.error_message),
        )


def _comparison_side_from_response(side_data):
    # type: (dict) -> ComparisonSide
    return ComparisonSide(
        file_type=str(side_data["file_type"]),
        source_url=side_data.get("source_url"),
        display_name=side_data.get("display_name"),
    )


def comparison_from_response(data):
    # type: (dict) -> Comparison
    return Comparison(
        identifier=str(data["identifier"]),
        left=_comparison_side_from_response(data["left"]),
        right=_comparison_side_from_response(data["right"]),
        public=data.get("public", False),
        creation_time=parse_datetime(data["creation_time"]),
        expiry_time=parse_datetime(data["expiry_time"])
        if "expiry_time" in data
        else None,
        ready=data.get("ready"),
        ready_time=parse_datetime(data["ready_time"]) if "ready_time" in data else None,
        failed=data.get("failed"),
        error_message=data.get("error_message"),
    )
